"""Benchmarks for graphql

Benchmarks are disabled (only executed as tests) by default in setup.cfg.
You can enable them with --benchmark-enable if your want to execute them.

E.g. in order to execute all the benchmarks with tox using Python 3.9::

    tox -e py39 -- -k benchmarks --benchmark-enable
"""
